#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import ctypes
import logging
from haystack.model import is_valid_address,getaddress,array2bytes,bytes2array,LoadableMembers,RangeValue,NotNull,CString,EVP_CIPHER_CTX_APP_DATA_PTR
import haystack.model
import ctypes_nss_generated as gen

log=logging.getLogger('ctypes_nss')

'''
http://mxr.mozilla.org/firefox/source/security/nss/lib/ssl/ssl.h
http://mxr.mozilla.org/firefox/source/security/nss/lib/ssl/sslt.h
http://mxr.mozilla.org/firefox/source/security/nss/lib/ssl/sslimpl.h#912
'''

class NSSStruct(LoadableMembers):
  ''' defines classRef '''
  pass

def pasteNSSStructOver(klass):
  model.pasteLoadableMemberMethodsOn(klass)
  klass.classRef = NSSStruct.classRef
  return klass

'''
 We are gonna load all generated structures into local __module__, adding
 LoadableMembers method classes and classRef on them.
 
 We just have to define expectedValues
'''

# replace c_char_p with our String handler
if type(gen.STRING) != type(CString):
  print 'STRING is not model.CString. Please correct ctypes_nss_geenrated with :' 
  print 'from model import CString' 
  print 'STRING = CString' 
  import sys
  sys.exit()

# set expected values 
gen.SSLCipherSuiteInfo.expectedValues={
  "cipherSuite": RangeValue(0,0x0100), # sslproto.h , ECC is 0xc00
  "authAlgorithm": RangeValue(0,4),
  "keaType": RangeValue(0,4),
  "symCipher": RangeValue(0,9),
  "macAlgorithm": RangeValue(0,4),
}

gen.SSLChannelInfo.expectedValues={
  'compressionMethod':RangeValue(0,1),
}

gen.SECItem.expectedValues={
  'type':RangeValue(0,15), # security/nss/lib/util/seccomon.h:64
}

gen.SECKEYPublicKey.expectedValues={
  'keyType':RangeValue(0,6), # security/nss/lib/cryptohi/keythi.h:189
}

#gen.NSSLOWKEYPublicKey.expectedValues={
#  'keyType':RangeValue(0,5), # security/nss/lib/softoken/legacydb/lowkeyti.h:123
#}

gen.CERTCertificate.expectedValues={ # XXX TODO CHECK VALUES security/nss/lib/certdb/certt.h
  'keyIDGenerated': [0,1], # ok
  'keyUsage': [RangeValue(0,gen.KU_ALL), # security/nss/lib/certdb/certt.h:570
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT ,
      RangeValue(gen.KU_NS_GOVT_APPROVED, gen.KU_ALL | gen.KU_NS_GOVT_APPROVED),
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT | gen.KU_NS_GOVT_APPROVED ] ,
  'rawKeyUsage': [RangeValue(0,gen.KU_ALL), # security/nss/lib/certdb/certt.h:570
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT ,
      RangeValue(gen.KU_NS_GOVT_APPROVED, gen.KU_ALL | gen.KU_NS_GOVT_APPROVED),
      gen.KU_KEY_AGREEMENT_OR_ENCIPHERMENT | gen.KU_NS_GOVT_APPROVED ] ,
  'keyUsagePresent': [0,1], # ok
  'nsCertType': [ RangeValue(0,0xff), # security/nss/lib/certdb/certt.h:459
      RangeValue(gen.EXT_KEY_USAGE_TIME_STAMP , gen.EXT_KEY_USAGE_TIME_STAMP | 0xff ), # not sure
      RangeValue(gen.EXT_KEY_USAGE_STATUS_RESPONDER  , gen.EXT_KEY_USAGE_STATUS_RESPONDER | 0xff ), # not sure
      RangeValue(gen.EXT_KEY_USAGE_TIME_STAMP | gen.EXT_KEY_USAGE_STATUS_RESPONDER  , 
                  gen.EXT_KEY_USAGE_TIME_STAMP | gen.EXT_KEY_USAGE_STATUS_RESPONDER | 0xff ) ] , # not sure
  'keepSession': [0,1], # ok
  'timeOK': [0,1], # ok
  'isperm': [0,1], # ok
  'istemp': [0,1], # ok
  'isRoot': [0,1], # ok
  'ownSlot': [0,1], # ok
  'referenceCount': RangeValue(1,0xffff), # no idea... should be positive for sure. more than 65535 ref ? naah
  'subjectName': NotNull, # a certificate must have one...
}

gen.CERTSignedCrl.expectedValues = {
  'isperm': [0,1], # ok
  'istemp': [0,1], # ok
  'referenceCount': RangeValue(1,0xffff), # no idea... should be positive for sure. more than 65535 ref ? naah
}

gen.CERTCertTrustStr.expectedValues = {
  #'sslFlags': [1<<i for i in range(0,11+1)] , # ok security/nss/lib/softoken/legacydb/pcertt.h:438
  'sslFlags': RangeValue(0, 1<<12) , # simpler
  'emailFlags': RangeValue(0, 1<<12) , # simpler
  'objectSigningFlags': RangeValue(0, 1<<12) , # simpler
}

import inspect,sys
# auto import gen.* into . ?
__my_module__ = sys.modules[__name__]
_loaded=0
for (name,klass) in inspect.getmembers(gen, inspect.isclass):
  if type(klass) == type(ctypes.Structure) and klass.__module__ == 'ctypes_nss_generated' :
    setattr(__my_module__, name, klass)
    _loaded+=1
log.debug('loaded %d C structs from NSS'%(_loaded))

''' Load all generateed classes to local classRef '''
NSSStruct.classRef=dict([ (ctypes.POINTER( klass), klass) for (name,klass) in inspect.getmembers(sys.modules[__name__], inspect.isclass) if klass.__module__ == __name__ or klass.__module__ == 'ctypes_nss_generated'])

''' Load all model classes and create a similar non-ctypes Python class  
  thoses will be used to translate non pickable ctypes into POPOs.
'''
for klass,typ in inspect.getmembers(sys.modules[__name__], inspect.isclass):
  if typ.__module__ == __name__:
    setattr(sys.modules[__name__], '%s_py'%(klass), type('%s_py'%(klass),(object,),{}) )

# copy classRef and methods, we have to wait after all class are loaded and in NSSStruct.classRef
for (name,klass) in inspect.getmembers(gen, inspect.isclass):
  if type(klass) == type(ctypes.Structure) and klass.__module__ == 'ctypes_nss_generated' :
    pasteNSSStructOver(klass)





def printSizeof(mini=-1):
  for (name,klass) in inspect.getmembers(sys.modules[__name__], inspect.isclass):
    if type(klass) == type(ctypes.Structure) and klass.__module__ == 'ctypes_nss_generated' :
      if ctypes.sizeof(klass) > mini:
        print '%s:'%name,ctypes.sizeof(klass)
  #print 'SSLCipherSuiteInfo:',ctypes.sizeof(SSLCipherSuiteInfo)
  #print 'SSLChannelInfo:',ctypes.sizeof(SSLChannelInfo)

if __name__ == '__main__':
  printSizeof()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import os,logging,sys, copy

import ctypes
from ctypes import cdll
from ctypes_openssh import AES_BLOCK_SIZE, ssh_aes_ctr_ctx
from ctypes_openssl import AES_KEY, EVP_AES_KEY #, BF_KEY, EVP_bf_key

from haystack import model

log=logging.getLogger('engine')

libopenssl=cdll.LoadLibrary('libssl.so')



class Engine:

  def decrypt(self,block):
    ''' decrypts '''
    bLen=len(block)
    if bLen % AES_BLOCK_SIZE:
      log.error("Sugar, why do you give me a block the wrong size: %d not modulo of %d"%(bLen, AES_BLOCK_SIZE))
      return None
    data=(ctypes.c_ubyte*bLen )()
    for i in range(0, bLen ):
      #print i, block[i] 
      data[i]=ord(block[i])
    # no way....
    return self._decrypt(data,bLen)
  
  def _decrypt(self,data,bLen):
    raise NotImplementedError


def myhex(bstr):
  s=''
  for el in bstr:
    s+='\\'+hex(ord(el))[1:]
  return s


class StatefulAES_CBC_Engine(Engine):
  def __init__(self, context  ):
    self.sync(context)
    self._AES_cbc=libopenssl.AES_cbc_encrypt
    log.debug('cipher:%s block_size: %d key_len: %d '%(context.name, context.block_size, context.key_len))
  
  def _decrypt(self, src, bLen):
    buf=(ctypes.c_ubyte*AES_BLOCK_SIZE)()
    dest=(ctypes.c_ubyte*bLen)()
    enc=ctypes.c_uint(0)  ## 0 is decrypt for inbound traffic
    ##log.debug('BEFORE %s'%( myhex(self.aes_key_ctx.getCounter())) )
    #void AES_cbc_encrypt(
    #      const unsigned char *in, unsigned char *out, const unsigned long length, 
    #           const AES_KEY *key, unsigned char ivec[AES_BLOCK_SIZE], const int enc
    #        	  )
    self._AES_cbc( ctypes.byref(src), ctypes.byref(dest), bLen, ctypes.byref(self.key), 
              ctypes.byref(self.iv), enc ) 
    ##log.debug('AFTER  %s'%( myhex(self.aes_key_ctx.getCounter())) )
    print self, repr(model.array2bytes(dest))
    return model.array2bytes(dest)
  
  def sync(self, context):
    ''' refresh the crypto state '''
    self.evp_aes_key = EVP_AES_KEY().fromPyObj(context.evpCtx.cipher_data) # 
    # we need nothing else
    self.key = self.evp_aes_key.ks
    #print self.key
    # copy counter content
    self.iv = model.bytes2array(context.evpCtx.iv, ctypes.c_ubyte)
    log.info('IV value is %s'%(myhex(context.evpCtx.iv)) )
    # TODO , check si les IV outbound sont les bons.


class StatefulAES_Ctr_Engine(Engine):
  #ctx->cipher->do_cipher(ctx,out,in,inl);
  # -> openssl.AES_ctr128_encrypt(&in,&out,length,&aes_key, ivecArray, ecount_bufArray, &num )
  #AES_encrypt(ivec, ecount_buf, key); # aes_key is struct with cnt, key is really AES_KEY->aes_ctx
  #AES_ctr128_inc(ivec); #ssh_Ctr128_inc semble etre different, mais paramiko le fait non ?
  def __init__(self, context ):
    self.sync(context)
    self._AES_ctr=libopenssl.AES_ctr128_encrypt
    log.debug('cipher:%s block_size: %d key_len: %d '%(context.name, context.block_size, context.key_len))
  
  def _decrypt(self,block, bLen):
    buf=(ctypes.c_ubyte*AES_BLOCK_SIZE)()
    dest=(ctypes.c_ubyte*bLen)()
    num=ctypes.c_uint()
    ##log.debug('BEFORE %s'%( myhex(self.aes_key_ctx.getCounter())) )
    #void AES_ctr128_encrypt(
    #      const unsigned char *in, unsigned char *out, const unsigned long length, 
    #           const AES_KEY *key, unsigned char ivec[AES_BLOCK_SIZE],     
    #        	  unsigned char ecount_buf[AES_BLOCK_SIZE],  unsigned int *num)
    # debug counter overflow
    ###last=self.aes_key_ctx.getCounter()[-1]
    ###before=self.getCounter()
    self._AES_ctr( ctypes.byref(block), ctypes.byref(dest), bLen, ctypes.byref(self.key), 
              ctypes.byref(self.counter), ctypes.byref(buf), ctypes.byref(num) ) 
    '''
    newlast=self.aes_key_ctx.getCounter()[-1]
    if newlast < last :
      log.warning('Counter has overflown')
      after=self.getCounter()
      log.warning('Before %s'%(before))
      log.warning('After  %s'%(after))
    '''
    ##log.debug('AFTER  %s'%( myhex(self.aes_key_ctx.getCounter())) )
    return model.array2bytes(dest)
  
  def sync(self, context):
    ''' refresh the crypto state '''
    self.aes_key_ctx = ssh_aes_ctr_ctx().fromPyObj(context.app_data)
    # we need nothing else
    self.key = self.aes_key_ctx.aes_ctx
    # copy counter content
    self.counter = self.aes_key_ctx.aes_counter
    log.info('Counter value is %s'%(myhex(self.aes_key_ctx.getCounter())) )

  def getCounter(self):
    return myhex(self.aes_key_ctx.getCounter())

  def incCounter(self):
    ctr=self.counter
    for i in range(len(ctr)-1,-1,-1):
      ctr[i] += 1
      if ctr[i] != 0:
        return
    
  def decCounter(self):
    ctr=self.counter
    for i in range(len(ctr)-1,-1,-1):
      old = ctr[i]         
      ctr[i] -= 1
      if old != 0: # underflow
        return


class StatefulBlowfish_CBC_Engine(Engine):
  def __init__(self, context  ):
    self.sync(context)
    self._BF_cbc=libopenssl.BF_cbc_encrypt
    log.debug('cipher:%s block_size: %d key_len: %d '%(context.name, context.block_size, context.key_len))
  
  def _decrypt(self, src, bLen):
    BF_ROUNDS	= 16
    BF_BLOCK = 8
    buf=(ctypes.c_ubyte*BF_BLOCK)()
    dest=(ctypes.c_ubyte*bLen)()
    enc=ctypes.c_uint(0)  ## 0 is decrypt for inbound traffic

    #void BF_cbc_encrypt(const unsigned char *in, unsigned char *out, long length,
    #	const BF_KEY *schedule, unsigned char *ivec, int enc);void AES_cbc_encrypt(
    self._BF_cbc( ctypes.byref(src), ctypes.byref(dest), bLen, ctypes.byref(self.key), 
              ctypes.byref(self.iv), enc ) 

    print self, repr(model.array2bytes(dest))
    return model.array2bytes(dest)
  
  def sync(self, context):
    ''' refresh the crypto state '''
    self.bf_key = BF_KEY().fromPyObj(context.evpCtx.cipher_data) # 
    # we need nothing else BF_KEY
    self.key = self.bf_key
    #print self.key
    # copy counter content
    self.iv = model.bytes2array(context.evpCtx.iv, ctypes.c_ubyte)
    log.info('IV value is %s'%(myhex(context.evpCtx.iv)) )
    # TODO , check si les IV outbound sont les bons.



CIPHERS = {
  "none": None, #(		SSH_CIPHER_NONE, 8, 0, 0, 0, EVP_enc_null ),
  "des": None, #(		SSH_CIPHER_DES, 8, 8, 0, 1, EVP_des_cbc ),
  "3des": None, #(		SSH_CIPHER_3DES, 8, 16, 0, 1, evp_ssh1_3des ),
  "blowfish": None, #(		SSH_CIPHER_BLOWFISH, 8, 32, 0, 1, evp_ssh1_bf ),
  "3des-cbc": None, #(		SSH_CIPHER_SSH2, 8, 24, 0, 1, EVP_des_ede3_cbc ),
  "blowfish-cbc": None, #(	SSH_CIPHER_SSH2, 8, 16, 0, 1, EVP_bf_cbc ),
  "cast128-cbc": None, #(	SSH_CIPHER_SSH2, 8, 16, 0, 1, EVP_cast5_cbc ),
  "arcfour": None, #(		SSH_CIPHER_SSH2, 8, 16, 0, 0, EVP_rc4 ),
  "arcfour128": None, #(		SSH_CIPHER_SSH2, 8, 16, 1536, 0, EVP_rc4 ),
  "arcfour256": None, #(		SSH_CIPHER_SSH2, 8, 32, 1536, 0, EVP_rc4 ),
  "aes128-cbc": StatefulAES_CBC_Engine, 
  "aes192-cbc": StatefulAES_CBC_Engine, 
  "aes256-cbc": StatefulAES_CBC_Engine, 
  "rijndael-cbc@lysator.liu.se": StatefulAES_CBC_Engine, 
  "aes128-ctr": StatefulAES_Ctr_Engine,
  "aes192-ctr": StatefulAES_Ctr_Engine,
  "aes256-ctr": StatefulAES_Ctr_Engine,
}



def testDecrypt():
  buf='?A\xb7\ru\xc9\x08\xe2em\x16\x06\x1a\x18\xfb\x805,\xd8\x1f\x11\xa3\x1b )G\xe2\r`\xfaw\x87\xef\xfa\xa7\x95\xe1\x84>\xe1\x90\xec\xe1\xfa\xe5\x1e\x9c\xe3'



def main(argv):
  logging.basicConfig(level=logging.INFO)
  logging.debug(argv)

  testDecrypt()
  return -1


if __name__ == "__main__":
  main(sys.argv[1:])





# lrucache.py -- a simple LRU (Least-Recently-Used) cache class

# Copyright 2004 Evan Prodromou <evan@bad.dynu.ca>
# Licensed under the Academic Free License 2.1

# arch-tag: LRU cache main module

"""a simple LRU (Least-Recently-Used) cache module

This module provides very simple LRU (Least-Recently-Used) cache
functionality.

An *in-memory cache* is useful for storing the results of an
'expensive' process (one that takes a lot of time or resources) for
later re-use. Typical examples are accessing data from the filesystem,
a database, or a network location. If you know you'll need to re-read
the data again, it can help to keep it in a cache.

You *can* use a Python dictionary as a cache for some purposes.
However, if the results you're caching are large, or you have a lot of
possible results, this can be impractical memory-wise.

An *LRU cache*, on the other hand, only keeps _some_ of the results in
memory, which keeps you from overusing resources. The cache is bounded
by a maximum size; if you try to add more values to the cache, it will
automatically discard the values that you haven't read or written to
in the longest time. In other words, the least-recently-used items are
discarded. [1]_

.. [1]: 'Discarded' here means 'removed from the cache'.

"""

from __future__ import generators
import time
from heapq import heappush, heappop, heapify

__version__ = "0.2"
__all__ = ['CacheKeyError', 'LRUCache', 'DEFAULT_SIZE']
__docformat__ = 'reStructuredText en'

DEFAULT_SIZE = 16
"""Default size of a new LRUCache object, if no 'size' argument is given."""

class CacheKeyError(KeyError):
    """Error raised when cache requests fail
    
    When a cache record is accessed which no longer exists (or never did),
    this error is raised. To avoid it, you may want to check for the existence
    of a cache record before reading or deleting it."""
    pass

class LRUCache(object):
    """Least-Recently-Used (LRU) cache.
    
    Instances of this class provide a least-recently-used (LRU) cache. They
    emulate a Python mapping type. You can use an LRU cache more or less like
    a Python dictionary, with the exception that objects you put into the
    cache may be discarded before you take them out.
    
    Some example usage::
	
    cache = LRUCache(32) # new cache
    cache['foo'] = get_file_contents('foo') # or whatever
	
    if 'foo' in cache: # if it's still in cache...
	    # use cached version
        contents = cache['foo']
    else:
	    # recalculate
        contents = get_file_contents('foo')
	    # store in cache for next time
        cache['foo'] = contents

    print cache.size # Maximum size
	
    print len(cache) # 0 <= len(cache) <= cache.size
	
    cache.size = 10 # Auto-shrink on size assignment
	
    for i in range(50): # note: larger than cache size
        cache[i] = i
	    
    if 0 not in cache: print 'Zero was discarded.'

    if 42 in cache:
        del cache[42] # Manual deletion

    for j in cache:   # iterate (in LRU order)
        print j, cache[j] # iterator produces keys, not values
    """
    
    class __Node(object):
        """Record of a cached value. Not for public consumption."""
        
        def __init__(self, key, obj, timestamp):
            object.__init__(self)
            self.key = key
            self.obj = obj
            self.atime = timestamp
            self.mtime = self.atime
    
        def __cmp__(self, other):
            return cmp(self.atime, other.atime)

        def __repr__(self):
            return "<%s %s => %s (%s)>" % \
                   (self.__class__, self.key, self.obj, \
                    time.asctime(time.localtime(self.atime)))

    def __init__(self, size=DEFAULT_SIZE):
        # Check arguments
        if size <= 0:
            raise ValueError, size
        elif type(size) is not type(0):
            raise TypeError, size
        object.__init__(self)	
        self.__heap = []
        self.__dict = {}
        self.size = size
        """Maximum size of the cache.
        If more than 'size' elements are added to the cache,
        the least-recently-used ones will be discarded."""
    
    def __len__(self):
        return len(self.__heap)
    
    def __contains__(self, key):
        return self.__dict.has_key(key)
    
    def __setitem__(self, key, obj):
        if self.__dict.has_key(key):
            node = self.__dict[key]
            node.obj = obj
            node.atime = time.time()
            node.mtime = node.atime
            heapify(self.__heap)
        else:
            # size may have been reset, so we loop
            while len(self.__heap) >= self.size:
                lru = heappop(self.__heap)
                del self.__dict[lru.key]
            node = self.__Node(key, obj, time.time())
            self.__dict[key] = node
            heappush(self.__heap, node)
    
    def __getitem__(self, key):
        if not self.__dict.has_key(key):
            raise CacheKeyError(key)
        else:
            node = self.__dict[key]
            node.atime = time.time()
            heapify(self.__heap)
            return node.obj
    
    def __delitem__(self, key):
        if not self.__dict.has_key(key):
            raise CacheKeyError(key)
        else:
            node = self.__dict[key]
            del self.__dict[key]
            self.__heap.remove(node)
            heapify(self.__heap)
            return node.obj

    def __iter__(self):
        copy = self.__heap[:]
        while len(copy) > 0:
            node = heappop(copy)
            yield node.key
        raise StopIteration

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        # automagically shrink heap on resize
        if name == 'size':
            while len(self.__heap) > value:
                lru = heappop(self.__heap)
                del self.__dict[lru.key]
        
    def __repr__(self):
        return "<%s (%d elements)>" % (str(self.__class__), len(self.__heap))

    def mtime(self, key):
        """Return the last modification time for the cache record with key.
        May be useful for cache instances where the stored values can get
        'stale', such as caching file or network resource contents."""
        if not self.__dict.has_key(key):
            raise CacheKeyError(key)
        else:
            node = self.__dict[key]
            return node.mtime

if __name__ == "__main__":
    cache = LRUCache(25)
    print cache
    for i in range(50):
        cache[i] = str(i)
    print cache
    if 46 in cache:
        del cache[46]
    print cache
    cache.size = 10
    print cache
    cache[46] = '46'
    print cache
    print len(cache)
    for c in cache:
        print c
    print cache
    print cache.mtime(46)
    for c in cache:
        print c
